{{
i2cKbMouse.spin COG device driver.

Purpose of this object is to provide communications services
between the PS2I2C attached keyboard and mouse device.

Functions of this code:

Interface between this object and applications is over I2C via
queues. A power of 2 sized circular queue is maintained for
each device. The device driver will place data into a queue and
increment the head pointer. The user will read data and
increment the tail pointer. The user must read the queue fast
enough so that old data is not destroyed. The user will know
if data is available when head value is not equal tail.

This driver is an I2C "master" and talks to the PS2 I2C "slave".
The terms master and slave are defined in the I2C specification.

Functions of the PS2I2C device:

The PS2I2C is a communications pipe. It multiplexes Keyboard
and Mouse data on the I2C interface protocol. The device can
be read and written by the master. The master writes to the
mouse address and the PS2I2C controller passes the message on
to the mouse. Messages are similarly sent to the keyboard.
The master reads from the PS2I2C device's input queue. The
PS2I2C device relies on interrupts for low latency data
collection.

The key item that makes reading the mouse reliable is what we
call "back pressure". The controller will assert the PS2 clock
line while I2C transactions happen. Without this back pressure
feature, the design would not be usable.
}}

#ifdef GUIDEMO
OBJ sx : "FullDuplexStub"
#else
#define USEKEYBD
'{
CON
   _clkmode  = xtal1 + pll16x
   _clkfreq  = 80_000_000

OBJ
'  sx : "FullDuplexSingleton"
  sx : "TV_Text"

PUB main | mx,my,mz,k

'  sx.start(31,30,0,115200)
  sx.start(20)

  waitcnt(clkfreq+cnt)
  sx.str(string($d,"i2cKbMouseCog test"))

  start(28,29)

  sx.out(0)
{
  repeat
    k := getkey
    'sx.hex(k,4)
    'sx.out(".")
    sx.out(k)
    'sx.out(" ")
'}

{
  sx.str(string($d,"Mouse Type "))
  sx.hex(present,2)
  sx.out($d)
  waitcnt(clkfreq/2+cnt)
'}
'{
  repeat
    sx.out(0)
{
    mx := abs_x
    my := abs_y
    mz := abs_z
    sx.out($d)
    sx.hex(mx,8)
    sx.out(" ")
    sx.hex(my,8)
    sx.out(" ")
    sx.hex(mz,8)
    sx.out(" ")
    sx.hex(buttons,2)
    sx.out(" ")
    if button(2)
      sx.out("C")
    else
      sx.out(" ")
    if button(1)
      sx.out("R")
    else
      sx.out(" ")
    if button(0)
      sx.out("L")
    else
      sx.out(" ")
    waitcnt(clkfreq/50+cnt)
'}
#ifdef USEKEYBD
    dump(khead,ktail,@kqueue,QSIZE,3)
    sx.str(string($d,"K "))
    sx.hex(par_head,4)
    sx.out(" ")
    sx.hex(par_tail,4)
    sx.out(" ")
    sx.hex(@par_keys+par_tail<<1,4)
    sx.out(" ")
    sx.hex(par_keys.word[par_tail],4)
    sx.out(" ")
    'sx.out(par_keys.word[par_tail] & $ff)

    if par_head <> par_tail
      k := getkey
      sx.out(k)
{
      sx.out(" ")
      sx.hex(par_head,4)
      sx.out(" ")
      sx.hex(par_tail,4)
      'repeat

      repeat mx from 0 to $f
        sx.hex(par_keys.word[mx],2)
        sx.out(" ")
}
'    sx.out($d)
'    sx.hex(@darray,4)
    sx.out($d)
    repeat mx from 0 to 7
      sx.hex(word[@darray][mx],4)
      sx.out(" ")

    sx.out($d)
    'sx.hex(@par_states,4)
    repeat mx from 0 to 7
      if mx == 4
        sx.out($d)
      sx.hex(par_states[mx],8)
      sx.out(" ")

#else
    dump(mhead,mtail,@mqueue,QSIZE,present)
#endif
    waitcnt(_clkfreq/5+cnt)
'}

PRI dump(dhead,dtail,buf,len,type) | n, mod

  if type < 3
    mod := 3
  else
    mod := 4

'  sx.out($d)
  sx.str(string("Type "))
  sx.dec(type)
  sx.str(string(" Q Head:Tail "))
'  sx.hex(dhead,4)
'  sx.out(":")
'  sx.hex(dtail,4)
'  sx.out(" ")
  sx.hex(buf,4)
  sx.out(" ")
  sx.hex(dhead+buf,4)
  sx.out(":")
  sx.hex(dtail+buf,4)
'  sx.out($d)

'{
  len--
  repeat n from 0 to len
    ifnot (n // (mod*4))
      sx.str(string($d))
      sx.hex(buf+n,3)
    ifnot (n // mod)
      sx.out(" ")
    sx.hex(byte[buf][n],2)

'  sx.out($d)
'  sx.out($d)
'}

'}
#endif

CON

  QSIZE = 1 << 7                ' power of 2 queue should be at least 8 but prefer 64

VAR     ' cog interface - must keep in this order
  byte  cog

  long  sclmsk, sdamsk          ' pins masks for interface
  long  wrcmd,  wrdata          ' write command mailbox interface
  long  khead,  ktail           ' keybd queue pointers
  long  mhead,  mtail           ' mouse queue pointers
  long  kqptr,  mqptr           ' pointers to queues
  long  kparm,  mparm           ' parameters for keyboard and mouse

DAT
  darray        byte 0 [16]      ' debug array
  kqueue        byte 0 [QSIZE]  ' keybd queue
  mqueue        byte 0 [QSIZE]  ' mouse queue

CON

  ACK           = 0             ' I2C Acknowledge
  NAK           = 1             ' I2C No Acknowledge
  RECV          = 1             ' I2C Direction Receive

  DEVICEID      = $80           ' user's device id
  MOUSE         = DEVICEID
  KEYBD         = MOUSE+2

  PS2RESET      = $FF           ' Common reset command byte
  PS2RESEND     = $FF           ' Common resend command byte
  PS2ACK        = $FA           ' Common acknowledge byte
  PS2DISABLE    = $F5           ' Common disable command byte
  PS2ENABLE     = $F4           ' Common enable command byte
  PS2RATE       = $F3           ' Used by mouse for ID. used by keybd for typerate
  PS2GETID      = $F2           ' Common get device ID command
  PS2MREMOTE    = $F0           ' Mouse set remote command
  PS2KSETSCAN   = $F0           ' Keyboard set scan code command
  PS2BATOK      = $AA           ' Common Basic Assurance Test OK byte
  PS2BATFAIL    = $FC           ' Common Basic Assurance Test Failure byte
  PS2NODATA     = $A5           ' Common no data available byte
  PS2SCALE221   = $E7
  PS2SCALE121   = $E6

  PS2SETLEDS    = $ED
  SCRLOCK       = 1
  NUMLOCK       = 2
  CAPLOCK       = 4

  delay         = 10
  RETRIES       = 5             ' number of reset retries before giving up on a device

'=============================================================================
' start the interface
'
PUB start(clkpin,datpin)
''
'' All lock-keys will be enabled, NumLock will be initially 'on',
'' and auto-repeat will be set to 15cps with a delay of .5s
'' Default keyboard startup. See startx for bitstream definitions.
''

  return startx(clkpin, datpin, %0_000_100, %01_01000)

PUB startx(clkpin, datpin, kblocks, kbrate)

'' Like start, but allows you to specify lock settings and auto-repeat
''
''   locks = lock setup
''           bit 6 disallows shift-alphas (case set soley by CapsLock)
''           bits 5..3 disallow toggle of NumLock/CapsLock/ScrollLock state
''           bits 2..0 specify initial state of NumLock/CapsLock/ScrollLock
''           (eg. %0_001_100 = disallow ScrollLock, NumLock initially 'on')
''
''   auto  = auto-repeat setup
''           bits 6..5 specify delay (0=.25s, 1=.5s, 2=.75s, 3=1s)
''           bits 4..0 specify repeat rate (0=30cps..31=2cps)
''           (eg %01_00000 = .5s delay, 30cps repeat)


  sclmsk := 1 << clkpin         ' setup pin masks
  sdamsk := 1 << datpin

  kqptr  := @kqueue             ' setup queue pointers
  mqptr  := @mqueue

  mparm  := @par_x              ' mouse parameter pointer

  par_present := 0              ' set zero so startup works

#ifdef USEKEYBD
  key_present := 0              ' set zero so startup works
  kparm := @par_tail
  long[kparm][2] := @keytable  ' present var has table at startup
  longmove(@par_keys,@clkpin,4) ' compatible with keyboard.spin
#endif

  wrcmd~~
  result := cog := cognew(@entry,@sclmsk)+1
  repeat while wrcmd

  MouseInit

PUB stop
  if cog
    cogstop(cog~ - 1)

PUB WriteByte(device, mdata)
{{
Simple method to write one byte at a time to the device
}}
{
  sx.str(string($d,"WriteByte "))
  sx.hex(device,2)
  sx.out(" ")
  sx.hex(data,2)
'}
  wrdata  := mdata
  wrcmd   := device
  repeat while wrcmd


PUB SendCmd(device, mdata)
{{
Simple method to write a command to device and wait for PS2ACK response.
}}
'{
  sx.str(string($d,"sendCmd "))
  sx.hex(mdata,2)
'}
  wrdata  := mdata
  wrcmd   := device
  repeat while wrcmd

  repeat
    result := read(device)
'{
    sx.out(" ")
    sx.hex(result,3)

'  sx.str(string($d,"sendCmd done",$d))
'}
    if result == PS2ACK
      quit

PUB ReadReady(device)
{{
Simple method to see if a byte is ready in the input queue
}}
  case device

    MOUSE:
      if mhead <> mtail
        return true

    KEYBD:
      if khead <> ktail
        return true

  return false

PUB ReadByte(device)
{{
Simple method to read one byte at a time from device queue
}}
  case device

    MOUSE:
      if mhead <> mtail
        result := mqueue[mtail] & $FF
        mtail  := ++mtail & constant(QSIZE-1)

    KEYBD:
      if khead <> ktail
        result := kqueue[ktail] & $FF
        ktail  := ++ktail & constant(QSIZE-1)


PUB read(device)
{{
Wait for data from device
}}
  case device

    MOUSE:
      repeat while mhead == mtail
      result := mqueue[mtail] & $FF
      mtail  := ++mtail & constant(QSIZE-1)

    KEYBD:
      repeat while khead == ktail
      result := kqueue[ktail] & $FF
      ktail  := ++ktail & constant(QSIZE-1)

PUB flush(device)

  case device

    MOUSE:
      mtail := mhead

    KEYBD:
      ktail := khead

PUB padZero(device)
{{
It is possible for packets to be misaligned at startup.
That is, a mouse packet can be interpreted badly.
To avoid this, pad the mouse queue with zero's before
starting the packet decoder. The packet decoder syncs
on the first byte with ($08 & byte) not zero.
}}
  case device

    MOUSE:
      repeat 2                  ' make some zero's in buffer
        mqptr[mhead] := 0
        mhead := ++mhead & constant(QSIZE-1)
      repeat
        mqptr[mhead] := 0
        mhead := ++mhead & constant(QSIZE-1)
        ifnot mhead & 3       ' using 3 mask means support only wheelmouse+
          quit
      mtail := mhead

    KEYBD:
      ktail := khead


CON

' =================================================================
' begin KEYBOARD code
' =================================================================

CON
'=============================================================================
' keyboard interface
'
#ifdef USEKEYBD
VAR
  long  par_tail        'key buffer tail        read/write      (19 contiguous longs)
  long  par_head        'key buffer head        read-only
  long  key_present     'keyboard present       read-only
  long  par_states[8]   'key states (256 bits)  read-only
  long  par_keys[8]     'key buffer (16 words)  read-only       (also used to pass initial parameters)

PUB getkeyhead
  return par_head
PUB getkeytail
  return par_tail

PUB kbpresent

'' Check if keyboard present - valid ~2s after start
'' returns t|f

  return key_present

PUB key : keycode

'' Get key (never waits)
'' returns key (0 if buffer empty)

  if par_tail <> par_head
    keycode := par_keys.word[par_tail]
    par_tail := ++par_tail & $F

PUB getkey : keycode

'' Get next key (may wait for keypress)
'' returns key

  repeat until (keycode := key)

PUB newkey : keycode

'' Clear buffer and get new key (always waits for keypress)
'' returns key

  par_tail := par_head
  keycode := getkey


PUB gotkey : truefalse

'' Check if any key in buffer
'' returns t|f

  truefalse := par_tail <> par_head

PUB clearkeys

'' Clear key buffer

  par_tail := par_head


PUB keystate(k) : state

'' Get the state of a particular key
'' returns t|f

  state := -(par_states[k >> 5] >> k & 1)

#endif


CON

' =================================================================
' begin MOUSE code
' =================================================================

VAR

  long  par_x                   'absolute x     read-only       (7 contiguous longs)
  long  par_y                   'absolute y     read-only
  long  par_z                   'absolute z     read-only
  long  par_buttons             'button states  read-only
  long  par_present             'mouse present  read-only

  long  oldx, oldy, oldz        'must be followed by parameters (10 contiguous longs)

  long  bx_min, by_min, bz_min  'min/max must be contiguous
  long  bx_max, by_max, bz_max
  long  bx_div, by_div, bz_div
  long  bx_acc, by_acc, bz_acc

  byte  mousetype


PUB MouseInit | n

'  sx.str(string($d,"MouseInit ... "))
  WriteByte(MOUSE,PS2RESET)

  repeat n from 0 to retries

    waitcnt(clkfreq/10+cnt)
    result := readByte(MOUSE)

    sx.out(" ")
    sx.hex(result,4)

    ifnot result
      next

    if result == PS2BATOK
      quit

    if result == PS2BATFAIL
      mouseType := 0
      par_present := 0
      return 0

  waitcnt(clkfreq/2+cnt)

  flush(MOUSE)
  SendCmd(MOUSE,PS2ENABLE)    ' just in case reset fails falsely

  flush(MOUSE)
  mouseType := MousePresent

  flush(MOUSE)
  padZero(MOUSE)                ' pad mouse packet for startup

  par_present := mousetype      ' enable mouse data collection

'{
  sx.str(string($d,"MouseType "))
  sx.dec(mouseType)
'}

  waitcnt(clkfreq/2+cnt)

PUB MousePresent : type
{{
'' Check if mouse present - valid ~2s after start
'' returns mouse type:
''
''   3 = five-button scrollwheel mouse
''   2 = three-button scrollwheel mouse
''   1 = two-button or three-button mouse
''   0 = no mouse connected
}}
  repeat
    sendCmd(MOUSE,PS2RATE)
    sendCmd(MOUSE,200)
    sendCmd(MOUSE,PS2RATE)
    sendCmd(MOUSE,100)
    sendCmd(MOUSE,PS2RATE)
    sendCmd(MOUSE,80)
    sendCmd(MOUSE,PS2GETID)
    mousetype := read(MOUSE)
    if mousetype == $FF
      mousetype := 1
      'quit
  while mousetype < 0 or mousetype > 8
  ifnot mousetype
    mousetype := 1              ' assume old style mouse
  return mousetype


PUB Present : type
{{
'' Check if mouse present - valid ~2s after start
'' returns mouse type:
''
''   3 = five-button scrollwheel mouse
''   2 = three-button scrollwheel mouse
''   1 = two-button or three-button mouse
''   0 = no mouse connected
}}
  return par_present

PUB button(b) : state

'' Get the state of a particular button
'' returns t|f

  state := -(par_buttons >> b & 1)

PUB buttons : states

'' Get the states of all buttons
'' returns buttons:
''
''   bit4 = right-side button ... not yet supported
''   bit3 = left-side button ... not yet supported
''   bit2 = center/scrollwheel button
''   bit1 = right button
''   bit0 = left button

  states := par_buttons

PUB abs_x

'' Get absolute-x

  return par_x


PUB abs_y

'' Get absolute-y

  return par_y


PUB abs_z

'' Get absolute-z (scrollwheel)

  return par_z

PUB delta_reset

'' Reset deltas

  oldx := par_x
  oldy := par_y
  oldz := par_z


PUB delta_x : mx | newx

'' Get delta-x

  newx := par_x
  mx := newx - oldx
  oldx := newx


PUB delta_y : my | newy

'' Get delta-y

  newy := par_y
  my := newy - oldy
  oldy := newy


PUB delta_z : mz | newz

'' Get delta-z (scrollwheel)

  newz := par_z
  mz := newz - oldz
  oldz := newz


PUB bound_limits(xmin, ymin, zmin, xmax, ymax, zmax) | i

'' Set bounding limits

  longmove(@bx_min, @xmin, 6)


PUB bound_scales(x_scale, y_scale, z_scale)

'' Set bounding scales (usually +/-1's, bigger values divide)

  longmove(@bx_div, @x_scale, 3)


PUB bound_preset(mx, my, mz) | i, d

'' Preset bound coordinates

  repeat i from 0 to 2
    d := ||bx_div[i]
    bx_acc[i] := (mx[i] - bx_min[i]) * d + d >> 1


PUB bound_x

'' Get bound-x

  return bound(0, delta_x)


PUB bound_y

'' Get bound-y

  return bound(1, delta_y)


PUB bound_z : z

'' Get bound-z

  return bound(2, delta_z)


PRI bound(i, delta) : b | d

  d := bx_div[i]
  b := bx_min[i] + (bx_acc[i] := bx_acc[i] + delta * (d < 0) | 1 #> 0 <# (bx_max[i] - bx_min[i] + 1) * ||d - 1) / ||d

CON

' =================================================================
' end of MOUSE code
' =================================================================


DAT     org
'============================================================
' i2c keyboard/mouse device COG driver
'============================================================
entry   ' get parameters then start command loop

tmp     mov     tmp,par

'------------------------------------------------------------
' initialize control/data pointers
'
_sclmsk         rdlong  $,tmp           ' scl mask
                add     tmp,#4
_sdamsk         rdlong  $,tmp           ' sda mask
                add     tmp,#4
_wrcmd          mov     $,tmp           ' _wrcmd is pointer to device id to write
                add     tmp,#4
_wrdata         mov     $,tmp           ' _wrdata is pointer to data
                add     tmp,#4
_khead          mov     $,tmp           ' pointer to khead
                add     tmp,#4
_ktail          mov     $,tmp           ' pointer to ktail
                add     tmp,#4
_mhead          mov     $,tmp           ' pointer to mhead
                add     tmp,#4
_mtail          mov     $,tmp           ' pointer to mtail
                add     tmp,#4
_kqueue         mov     $,tmp           ' pointer to kqueue address 0
                add     tmp,#4
_mqueue         mov     $,tmp           ' pointer to mqueue address 0
                add     tmp,#4
_kparm          mov     $,tmp           ' pointer to keyboard parms
                add     tmp,#4
_mparm          mov     $,tmp           ' pointer to mouse parms


#ifdef USEKEYBD
'------------------------------------------------------------
' initialize keyboard pointers
'
_kptr           rdlong  $,_kparm
_kbtail         mov     $,_kptr         ' keybd tail pointer
                add     _kptr,#4
_kbhead         mov     $,_kptr         ' keybd head pointer
                add     _kptr,#4
_kbpres         mov     $,_kptr         ' pointer to keybd present
_kbtable        rdlong  $,_kptr         ' keybd present variable = keytable at startup
                add     _kptr,#4
_kbstates       mov     $,_kptr         ' keybd states
                add     _kptr,#32       ' keybd states uses 8 longs
_kbkeyqp        mov     $,_kptr         ' key buffer queue pointer

                add     _kptr,#8        ' keybd autorate parameter
                rdlong  _locks,_kptr    ' get parameter
                add     _kptr,#8        ' keybd locks parameter
                rdlong  _auto,_kptr     ' get parameter

                mov     _kbpres,#0      ' no keyboard at startup
#endif

'------------------------------------------------------------
' initialize mouse pointers
'
_mptr           rdlong  $,_mparm        ' get mouse info address
_msx            mov     $,_mptr         ' pointer to mouse x
                add     _mptr,#4
_msy            mov     $,_mptr         ' pointer to mouse y
                add     _mptr,#4
_msz            mov     $,_mptr         ' pointer to mouse z
                add     _mptr,#4
_msbut          mov     $,_mptr         ' pointer to mouse buttons
                add     _mptr,#4
_mspres         mov     $,_mptr         ' pointer to mouse present

                call    #i2c_init

'------------------------------------------------------------
' permiscuous configuration protocol must pass before continuing
'
doPCP
                mov     i2cDev,#DEVICEID' assign user's device id
                or      i2cDev,#1       ' always a read
                call    #_write         ' send data
                call    #pdelay
        if_c    jmp     #doPCP          ' if ack ok, carry is clear

                mov     wcmd,#0         ' cog startup finished
                wrlong  wcmd,_wrcmd

'------------------------------------------------------------
' main command loop. poll device before checking for data write.
'
cmdloop
                rdlong  wcmd,_wrcmd wz
        if_nz   call    #writeDevice ' if wcmd <> 0, write wcmd device

                call    #pdelay
                call    #pollDevices    ' poll devices for data

                call    #updateMouse
#ifdef USEKEYBD
                call    #updateKeybd
#endif
                jmp     #cmdloop
'
'------------------------------------------------------------

pdelay
                mov     ptime,#$100     ' $100 << 8 gives enough delay so next poll skips
                shl     ptime,#8        ' beginning of PS2 packet. $110 << 8 lags but works.
                add     ptime,cnt
                waitcnt ptime,ptime     ' wait a while before poll
pdelay_ret      ret

ptime   long    0

wcmd    long    0
ndx     long    0
sdata   long    0

i2cAck  long    0
i2cDev  long    0
i2cData long    0
i2cPlen long    0
i2cPckt long    0

head    long    0
tail    long    0
tp      long    0
byt     long    0
butn    long    0
mtype   long    0
ktype   long    0

bext    long    $ffff_ff80
bmsk    long    $7f

#ifdef USEKEYBD

dlsb    long    1 << 9

stat    long    8
data    long    0
x       long    0
y       long    0
t       long    0

_head    long    0       'write-only
_present long    0       'write-only
_states  long    0 [8]   'write-only
_locks   long    0
_auto    long    0

'------------------------------------------------------------------------------
' handle keybd data
'
'
'
' Get scancode
'
updateKeybd             'receive byte from keyboard

                        test    stat,#8         wc      'if reset flag, transmit reset command
        if_c            mov     stat,#0                 ' same functionality ask keyboad.spin
        if_c            mov     data,#PS2RESET
        if_c            call    #transmit

:same                   'call  #receive                 ' this is receive
                        rdlong  head,_khead             ' except it's cooperatively tasked
                        rdlong  tail,_ktail
                        cmp     head,tail wz            ' do we have new queue value?
        if_e            jmp     #updateKeybd_ret        ' nothing to do. exit.

                        rdlong  tp,_kqueue              ' get i2c queue pointer
                        add     tp,tail                 ' get queue+tail
                        rdbyte  data,tp                 ' read byte from i2c queue

                        add     tail,#1
                        and     tail,#QSIZE-1           ' roll i2c rx tail
                        wrlong  tail,_ktail

                        cmp     data,#$83+1     wc      'scancode?

        if_nc           cmp     data,#$AA       wz      'powerup/reset?
        if_nc_and_z     jmp     #configure

        if_nc           cmp     data,#$E0       wz      'extended?
        if_nc_and_z     or      stat,#1
        if_nc_and_z     jmp     #updateKeybd_ret        ' back to updateKeybd is :same

        if_nc           cmp     data,#$F0       wz      'released?
        if_nc_and_z     or      stat,#2
        if_nc_and_z     jmp     #updateKeybd_ret        ' back to updateKeybd is :same

        if_nc           jmp     #newcode                'unknown, ignore. nothing to do. exit.
'
'
' Translate scancode and enter into buffer
'
                        'wrbyte  data,#$60               ' debug only

                        test    stat,#1         wc      'lookup code with extended flag
                        rcl     data,#1
                        call    #look

                        'wrbyte  data,#$62               ' debug only

                        cmp     data,#0         wz      'if unknown, ignore
        if_z            jmp     #newcode

                        mov     t,_states+6             'remember lock keys in _states

                        mov     x,data                  'set/clear key bit in _states
                        shr     x,#5
                        add     x,#_states
                        movd    :reg,x
                        mov     y,#1
                        shl     y,data
                        test    stat,#2         wc
:reg                    muxnc   0,y

        if_nc           cmpsub  data,#$F0       wc      'if released or shift/ctrl/alt/win, done
        if_c            jmp     #update

                        mov     y,_states+7             'get shift/ctrl/alt/win bit pairs
                        shr     y,#16

                        cmpsub  data,#$E0       wc      'translate keypad, considering numlock
        if_c            test    _locks,#%100    wz
        if_c_and_z      add     data,#@keypad1-@keytable       ' TBD check this
        if_c_and_nz     add     data,#@keypad2-@keytable       ' TBD check this
        if_c            call    #look
        if_c            jmp     #:flags

                        cmpsub  data,#$DD       wc      'handle scrlock/capslock/numlock
        if_c            mov     x,#%001_000
        if_c            shl     x,data
        if_c            andn    x,_locks
        if_c            shr     x,#3
        if_c            shr     t,#29                   'ignore auto-repeat
        if_c            andn    x,t             wz
        if_c            xor     _locks,x
        if_c            add     data,#$DD
        if_c_and_nz     or      stat,#4                 'if change, set configure flag to update leds

                        test    y,#%11          wz      'get shift into nz

        if_nz           cmp     data,#$60+1     wc      'check shift1
        if_nz_and_c     cmpsub  data,#$5B       wc
        if_nz_and_c     add     data,#@shift1-@keytable
        if_nz_and_c     call    #look
        if_nz_and_c     andn    y,#%11

        if_nz           cmp     data,#$3D+1     wc      'check shift2
        if_nz_and_c     cmpsub  data,#$27       wc
        if_nz_and_c     add     data,#@shift2-@keytable
        if_nz_and_c     call    #look
        if_nz_and_c     andn    y,#%11

                        test    _locks,#%010    wc      'check shift-alpha, considering capslock
                        muxnc   :shift,#$20
                        test    _locks,#$40     wc
        if_nz_and_nc    xor     :shift,#$20
                        cmp     data,#"z"+1     wc
        if_c            cmpsub  data,#"a"       wc
:shift  if_c            add     data,#"A"
        if_c            andn    y,#%11

:flags                  ror     data,#8                 'add shift/ctrl/alt/win flags
                        mov     x,#4                    '+$100 if shift
:loop                   test    y,#%11          wz      '+$200 if ctrl
                        shr     y,#2                    '+$400 if alt
        if_nz           or      data,#1                 '+$800 if win
                        ror     data,#1
                        djnz    x,#:loop
                        rol     data,#12

                        rdlong  x,_kbtail               'if room in buffer and key valid, enter
                        sub     x,#1
                        and     x,#$F
                        cmp     x,_head         wz
        if_nz           test    data,#$FF       wz
        if_nz           mov     x,_kbtail
        if_nz           add     x,#11*4
        if_nz           add     x,_head
        if_nz           add     x,_head
        if_nz           wrword  data,x
        if_nz           add     _head,#1
        if_nz           and     _head,#$F

                        test    stat,#4         wc      'if not configure flag, done
        if_nc           jmp     #update                 'else configure to update leds

'
'
' Configure keyboard
'
configure
                        mov     data,#$F3               'set keyboard auto-repeat
                        call    #transmit
                        mov     data,_auto
                        and     data,#%11_11111
                        wrword  data,#$6C
                        call    #transmit

                        mov     data,#$ED               'set keyboard lock-leds
                        call    #transmit
                        mov     data,_locks
                        rev     data,#-3 & $1F
                        test    data,#%100      wc
                        rcl     data,#1
                        and     data,#%111
                        wrword  data,#$6E
                        call    #transmit

                        mov     x,_locks                'insert locks into _states
                        and     x,#%111
                        shl     _states+7,#3
                        or      _states+7,x
                        ror     _states+7,#3

                        mov     _present,#1             'set _present
'
' update the parameters
' using original names to minimize changes in keybd driver
'
update
                        movd    :par,#_head             'update output parameters _head/_present/_states[8]
                        mov     x,_kbtail
                        add     x,#1*4
                        mov     y,#1+1+8
:par                    wrlong  0,x
                        add     :par,dlsb
                        add     x,#4
                        djnz    y,#:par
'
' newcode needs stat cleared
' using original names to minimize changes in keybd driver
'
newcode                 mov     stat,#0

updateKeybd_ret ret


'------------------------------------------------------------
' lookup keycode
' using original names to minimize changes in keybd driver
'
look            mov     tmp,_kbtable
                add     tmp,data                ' lookup keytable entry
                rdbyte  data,tmp

look_ret        ret


'------------------------------------------------------------
' send byte to keyboard
' using original names to minimize changes in keybd driver
'
transmit
                mov     i2cDev,#KEYBD           ' user's device id
                andn    i2cDev,#1               ' always a write
                mov     i2cData,data            ' user's data
                call    #i2c_wrbyte             ' send data
                call    #pdelay
transmit_ret    ret

#endif

'------------------------------------------------------------
' update mouse variables
'
updateMouse
        rdlong  mtype,_mspres
        tjz     mtype,#updateMouse_ret  ' no mouse yet

        mov     mbytes,#4
        cmp     mtype,#3 wc
if_b    mov     mbytes,#3

        rdlong  head,_mhead
        rdlong  tail,_mtail
        cmp     head,tail wz            ' head becomes length. tail stays same.
if_e    jmp     #updateMouse_ret        ' nothing to do

        mov     ndx,head
        sub     ndx,tail
        abs     ndx,ndx
        cmp     ndx,mbytes wc
if_b    jmp     #updateMouse_ret        ' if we don't have enough bytes, exit

        rdlong  tp,_mqueue
        add     tp,tail

:lp                                     ' search for packet start
        rdbyte  byt,tp
        add     tp,#1
        add     tail,#1

'
' The older two button mouse creates lots of overflows.
' Such packets should be discarded if 2 button mode is enabled.
' The practicallity of using such a device is very limited here.
'
' These exceptions allow us to sync up with the beginning of the packet.
' After the beginning is detected, we rely on the mouse to always deliver
' correct packets. If packets are bad, some correction is attempted
' by the sync up code, but recovery is not always 100% correct.
' This can result in mouse cursor jumping depending on number of errors.
'

{       ' uncomment to enable this check
        mov     tmp,byt
        and     tmp,#$0F wz             ' dis-allowing 3 button clicks
        cmp     tmp,#$0F wz             ' next
if_z    jmp     #:lp                    ' next
'}

        test    byt,#$40 wz             ' dis-allowing overflows
if_nz   jmp     #:lp                    ' next
        test    byt,#$80 wz             ' dis-allowing overflows
if_nz   jmp     #:lp                    ' next

        test    byt,#$8  wz             ' check for start packet
if_z    jmp     #:lp

        mov     butn,byt                ' save for sign checks
        wrlong  byt,_msbut              ' first byte is button

        rdbyte  byt,tp                  ' read byte from queue
        add     tp,#1
        add     tail,#1

        rdlong  tmp,_msx                ' read modify write value
        test    butn,#$10 wc            ' $10 is dx sign
if_c    or      byt,bext
if_nc   and     byt,bmsk
        add     tmp,byt
        wrlong  tmp,_msx                ' next byte is x

        rdbyte  byt,tp                  ' read byte from queue
        add     tp,#1
        add     tail,#1

        rdlong  tmp,_msy                ' read modify write value
        test    butn,#$20 wc            ' $20 is dy sign
if_c    or      byt,bext
if_nc   and     byt,bmsk
        add     tmp,byt
        wrlong  tmp,_msy                ' next byte is y

        cmp     mtype,#3 wc             ' mouse < 3 packet is but,x,y
if_b    jmp     #updateMouse_next       ' if mouse type >= 3, get z

        rdbyte  byt,tp                  ' read byte from queue
        add     tp,#1
        add     tail,#1

        rdlong  tmp,_msz                ' read modify write value
        test    tmp,#$80 wc
if_c    or      byt,bext
if_nc   and     byt,bmsk
        add     byt,byt
        wrlong  tmp,_msz                ' next byte is z

updateMouse_next
        and     tail,#QSIZE-1
        wrlong  tail,_mtail             ' update tail

        jmp     #updateMouse_ret
updateMouse_done

        rdlong  head,_mhead
        rdlong  head,_mtail

updateMouse_ret ret

mbytes  long    0


'------------------------------------------------------------
' write device
'
writeDevice
        rdlong  sdata,_wrdata
        mov     i2cDev,wcmd             ' user's device id
        call    #writeDeviceSub
        mov     ptime,#$100             ' give device a chance to respond
        shl     ptime,#11               ' before the read routine crushes clock.
        add     ptime,cnt
        waitcnt ptime,ptime             ' wait a while before poll
        mov     wcmd,#0                 ' write device finished
        wrlong  wcmd,_wrcmd
        jmp     #cmdloop
writeDevice_ret ret

writeDeviceSub
        andn    i2cDev,#1               ' always a write
        mov     i2cData,sdata           ' user's data
        call    #i2c_wrbyte             ' send data
writeDeviceSub_ret ret

'------------------------------------------------------------
' poll devices
'
pollDevices
        add     devpoll,#1
        test    devpoll,#$7 wz
if_z    jmp     #pollKeybd

        mov     i2cDev,#MOUSE           ' check MOUSE
        call    #i2c_chkbyte
if_c    jmp     #pollDevices_ret
        or      i2cAck,_sdamsk          ' just one byte
        call    #_read                  ' read byte
        call    #_stop
        call    #queueMouse
        jmp     #pollDevices_ret
        call    #pdelay
pollKeybd
        call    #kbreceive
pollDevices_ret ret

devpoll long    0

kbpoll
        mov     i2cDev,#KEYBD           ' check KEYBD
        mov     sdata,#$FA
        call    #writeDeviceSub
        call    #pdelay
        call    #kbreceive              ' receive the data if any
kbpoll_ret      ret

kbreceive
        mov     i2cDev,#KEYBD           ' check KEYBD
        call    #i2c_chkbyte
if_c    jmp     #kbreceive_ret
        or      i2cAck,_sdamsk          ' just one byte
        call    #_read                  ' read byte
        call    #_stop
        call    #queueKeybd
kbreceive_ret   ret

'------------------------------------------------------------
' queue data to mouse
'
queueMouse
        rdlong  tmp,_mqueue             ' get queue start
        rdlong  ndx,_mhead              ' set position
        add     tmp,ndx
        wrbyte  i2cData,tmp             ' write byte
        add     ndx,#1                  ' add to queue head
        and     ndx,#QSIZE-1            ' it's circular
        wrlong  ndx,_mhead
queueMouse_ret  ret

'------------------------------------------------------------
' queue data to keybd
'
queueKeybd
        rdlong  tmp,_kqueue             ' get queue start
        rdlong  ndx,_khead              ' set position
        add     tmp,ndx
        wrbyte  i2cData,tmp             ' write byte
        add     ndx,#1                  ' add to queue head
        and     ndx,#QSIZE-1            ' it's circular
        wrlong  ndx,_khead
queueKeybd_ret  ret

'------------------------------------------------------------
' initializes i2c devices to known state
' sends to address $ff which should get no ack
'
i2c_init
        or      outa,_sclmsk
        or      dira,_sclmsk
        andn    dira,_sdamsk
        mov     ndx,#9
:loop   andn    outa,_sclmsk
        or      outa,_sclmsk
        test    _sdamsk,ina wc          ' test for sda high
if_nc   djnz    ndx,#:loop              ' loop if not high
i2c_init_ret    ret

'------------------------------------------------------------
' chkbyte
'
' i2cDev  - device id
' carry set if no data available
' is device busy ? does device have data for us ?
'
i2c_chkbyte
        call    #_start
        mov     i2cData,i2cDev
        or      i2cData,#1               ' it's a read transaction
        call    #_write                  ' get device attention
        test    i2cAck,_sdamsk wc        ' if nack, return carry set
if_c    call    #_stop
i2c_chkbyte_ret  ret


'------------------------------------------------------------
' wrbyte
'
' i2cDev  - device id
' i2cData - data to write
'
i2c_wrbyte
        mov     sdata,i2cData
        call    #_wait
if_nc   mov     i2cData,sdata
if_nc   call    #_write
if_nc   call    #_stop
i2c_wrbyte_ret  ret

'------------------------------------------------------------
' wait
'
' i2cDev  - device id
' i2cAck  - cleared if device is responding
' carry clear if device responding
'
_wait
        call    #_start
        call    #i2c_delay
        andn    i2cDev,#1               ' it's a write transaction
        mov     i2cData,i2cDev
        call    #_write
        test    i2cAck,_sdamsk wc        ' wait for dev not busy
if_c    call    #_stop
if_c    jmp     #_wait
_wait_ret    ret

'------------------------------------------------------------
' sets start state
' start is SCL high with SDA High->Low
'
_start
        or      outa,_sclmsk            ' scl high
        or      outa,_sdamsk            ' sda high
        or      dira,_sdamsk            ' turn on sda
        or      dira,_sclmsk            ' turn on scl
        call    #i2c_delay
        andn    outa,_sdamsk            ' turn off sda
        call    #i2c_delay
        andn    outa,_sclmsk            ' turn off scl
        call    #_startdelay
_start_ret   ret

'------------------------------------------------------------
' sets stop state
' stop  is SCL high with SDA Low->High
'
_stop
        call    #_stopdelay
        or      outa,_sclmsk            ' scl high
        call    #i2c_delay
        or      outa,_sdamsk            ' sda high
        call    #i2c_delay
        andn    dira,_sclmsk            ' turn off scl
        andn    dira,_sdamsk            ' turn off sda
        call    #_stopdelay
_stop_ret    ret

'------------------------------------------------------------
' write data to i2c bus
' i2cData - data to write
'
_write
        call    #i2c_delay
        call    #i2c_delay
        ror     i2cData,#8              ' use shl wc to set bit
        mov     ndx,#8
:loop
        call    #i2c_delay              ' scl starts low
        shl     i2cData,#1 wc           ' set c based on msb at shl
        muxc    outa,_sdamsk            ' copy c bit to outa using mask
        call    #i2c_delay
        or      outa,_sclmsk            ' clock out data
        call    #i2c_delay
        call    #i2c_delay
        andn    outa,_sclmsk            ' scl low
        djnz    ndx,#:loop              ' next bit

        call    #i2c_delay
        andn    dira,_sdamsk            ' release sda to check for ack
        call    #i2c_delay
        or      outa,_sclmsk            ' scl was low - make high
        call    #i2c_delay

        mov     i2cAck,ina              ' get sda state
        and     i2cAck,_sdamsk          ' set ack flag ... non-zero is true

        call    #i2c_delay
        andn    outa,_sclmsk            ' scl ends low

        call    #i2c_delay
        andn    outa,_sdamsk            ' sda ends low
        or      dira,_sdamsk            ' sda ends driven
_write_ret   ret

'------------------------------------------------------------
' read data from i2c bus
' i2cAck  - ack value _sdamsk
' i2cData - data read
'
_read
        call    #i2c_delay
        call    #i2c_delay
        mov     i2cData,#0              ' start with 0
        andn    dira,_sdamsk            ' sda set to input
        mov     ndx,#8
:loop
        call    #i2c_delay              ' scl starts low

        or      outa,_sclmsk            ' scl goes hight to get bit
        call    #i2c_delay

        shl     i2cData,#1              ' addc does both muxc and shl ?
        test    _sdamsk,ina wc          ' get sda state
        muxc    i2cData,#1
        call    #i2c_delay

        andn    outa,_sclmsk            ' scl starts low
        call    #i2c_delay
        djnz    ndx,#:loop

        or      outa,i2cAck
        or      dira,_sdamsk            ' set the ack if required
        call    #i2c_delay
        or      outa,_sclmsk            ' scl goes hight to get bit
        call    #i2c_delay

        andn    outa,_sclmsk            ' scl ends low
        call    #i2c_delay
        andn    outa,_sdamsk            ' sda ends low
_read_ret    ret

'------------------------------------------------------------
' delay at start
'
_startdelay
        mov     scnt,#5
:loop   call    #i2c_delay
        djnz    scnt,#:loop
_startdelay_ret ret

'------------------------------------------------------------
' delay at stop
'
_stopdelay
        mov     scnt,#2
:loop   call    #i2c_delay
        djnz    scnt,#:loop
_stopdelay_ret ret

scnt    long    0

'------------------------------------------------------------
' delay between changing bit states
'
i2c_delay
        mov     dtime,dper
        add     dtime,cnt
        waitcnt dtime,dtime nr
i2c_delay_ret   ret     ' 10 x 50ns = 500ns

dtime   long    0
dper    long    $100     ' bad if < $a0. $100 is conservative.
'               $3b8    ' poll per ms

        fit     $1ef
'
'============================================================

#ifndef USEKEYBD
keytable                word    $0000   '00
keypad1                 byte    0
keypad2                 byte    0
shift1                  byte    0
shift2                  byte    0
#endif

#ifdef USEKEYBD
dat
'
'
' Lookup table
'                               ascii   scan    extkey  regkey  ()=keypad
'
keytable                word    $0000   '00
                        word    $00D8   '01             F9
                        word    $0000   '02
                        word    $00D4   '03             F5
                        word    $00D2   '04             F3
                        word    $00D0   '05             F1
                        word    $00D1   '06             F2
                        word    $00DB   '07             F12
                        word    $0000   '08
                        word    $00D9   '09             F10
                        word    $00D7   '0A             F8
                        word    $00D5   '0B             F6
                        word    $00D3   '0C             F4
                        word    $0009   '0D             Tab
                        word    $0060   '0E             `
                        word    $0000   '0F
                        word    $0000   '10
                        word    $F5F4   '11     Alt-R   Alt-L
                        word    $00F0   '12             Shift-L
                        word    $0000   '13
                        word    $F3F2   '14     Ctrl-R  Ctrl-L
                        word    $0071   '15             q
                        word    $0031   '16             1
                        word    $0000   '17
                        word    $0000   '18
                        word    $0000   '19
                        word    $007A   '1A             z
                        word    $0073   '1B             s
                        word    $0061   '1C             a
                        word    $0077   '1D             w
                        word    $0032   '1E             2
                        word    $F600   '1F     Win-L
                        word    $0000   '20
                        word    $0063   '21             c
                        word    $0078   '22             x
                        word    $0064   '23             d
                        word    $0065   '24             e
                        word    $0034   '25             4
                        word    $0033   '26             3
                        word    $F700   '27     Win-R
                        word    $0000   '28
                        word    $0020   '29             Space
                        word    $0076   '2A             v
                        word    $0066   '2B             f
                        word    $0074   '2C             t
                        word    $0072   '2D             r
                        word    $0035   '2E             5
                        word    $CC00   '2F     Apps
                        word    $0000   '30
                        word    $006E   '31             n
                        word    $0062   '32             b
                        word    $0068   '33             h
                        word    $0067   '34             g
                        word    $0079   '35             y
                        word    $0036   '36             6
                        word    $CD00   '37     Power
                        word    $0000   '38
                        word    $0000   '39
                        word    $006D   '3A             m
                        word    $006A   '3B             j
                        word    $0075   '3C             u
                        word    $0037   '3D             7
                        word    $0038   '3E             8
                        word    $CE00   '3F     Sleep
                        word    $0000   '40
                        word    $002C   '41             ,
                        word    $006B   '42             k
                        word    $0069   '43             i
                        word    $006F   '44             o
                        word    $0030   '45             0
                        word    $0039   '46             9
                        word    $0000   '47
                        word    $0000   '48
                        word    $002E   '49             .
                        word    $EF2F   '4A     (/)     /
                        word    $006C   '4B             l
                        word    $003B   '4C             ;
                        word    $0070   '4D             p
                        word    $002D   '4E             -
                        word    $0000   '4F
                        word    $0000   '50
                        word    $0000   '51
                        word    $0027   '52             '
                        word    $0000   '53
                        word    $005B   '54             [
                        word    $003D   '55             =
                        word    $0000   '56
                        word    $0000   '57
                        word    $00DE   '58             CapsLock
                        word    $00F1   '59             Shift-R
                        word    $EB0D   '5A     (Enter) Enter
                        word    $005D   '5B             ]
                        word    $0000   '5C
                        word    $005C   '5D             \
                        word    $CF00   '5E     WakeUp
                        word    $0000   '5F
                        word    $0000   '60
                        word    $0000   '61
                        word    $0000   '62
                        word    $0000   '63
                        word    $0000   '64
                        word    $0000   '65
                        word    $00C8   '66             BackSpace
                        word    $0000   '67
                        word    $0000   '68
                        word    $C5E1   '69     End     (1)
                        word    $0000   '6A
                        word    $C0E4   '6B     Left    (4)
                        word    $C4E7   '6C     Home    (7)
                        word    $0000   '6D
                        word    $0000   '6E
                        word    $0000   '6F
                        word    $CAE0   '70     Insert  (0)
                        word    $C9EA   '71     Delete  (.)
                        word    $C3E2   '72     Down    (2)
                        word    $00E5   '73             (5)
                        word    $C1E6   '74     Right   (6)
                        word    $C2E8   '75     Up      (8)
                        word    $00CB   '76             Esc
                        word    $00DF   '77             NumLock
                        word    $00DA   '78             F11
                        word    $00EC   '79             (+)
                        word    $C7E3   '7A     PageDn  (3)
                        word    $00ED   '7B             (-)
                        word    $DCEE   '7C     PrScr   (*)
                        word    $C6E9   '7D     PageUp  (9)
                        word    $00DD   '7E             ScrLock
                        word    $0000   '7F
                        word    $0000   '80
                        word    $0000   '81
                        word    $0000   '82
                        word    $00D6   '83             F7

keypad1                 byte    $CA, $C5, $C3, $C7, $C0, 0, $C1, $C4, $C2, $C6, $C9, $0D, "+-*/"

keypad2                 byte    "0123456789.", $0D, "+-*/"

shift1                  byte    "{|}", 0, 0, "~"

shift2                  byte    $22, 0, 0, 0, 0, "<_>?)!@#$%^&*(", 0, ":", 0, "+"
''
''
''      _________
''      Key Codes
''
''      00..DF  = keypress and keystate
''      E0..FF  = keystate only
''
''
''      09      Tab
''      0D      Enter
''      20      Space
''      21      !
''      22      "
''      23      #
''      24      $
''      25      %
''      26      &
''      27      '
''      28      (
''      29      )
''      2A      *
''      2B      +
''      2C      ,
''      2D      -
''      2E      .
''      2F      /
''      30      0..9
''      3A      :
''      3B      ;
''      3C      <
''      3D      =
''      3E      >
''      3F      ?
''      40      @
''      41..5A  A..Z
''      5B      [
''      5C      \
''      5D      ]
''      5E      ^
''      5F      _
''      60      `
''      61..7A  a..z
''      7B      {
''      7C      |
''      7D      }
''      7E      ~
''
''      80-BF   (future international character support)
''
''      C0      Left Arrow
''      C1      Right Arrow
''      C2      Up Arrow
''      C3      Down Arrow
''      C4      Home
''      C5      End
''      C6      Page Up
''      C7      Page Down
''      C8      Backspace
''      C9      Delete
''      CA      Insert
''      CB      Esc
''      CC      Apps
''      CD      Power
''      CE      Sleep
''      CF      Wakeup
''
''      D0..DB  F1..F12
''      DC      Print Screen
''      DD      Scroll Lock
''      DE      Caps Lock
''      DF      Num Lock
''
''      E0..E9  Keypad 0..9
''      EA      Keypad .
''      EB      Keypad Enter
''      EC      Keypad +
''      ED      Keypad -
''      EE      Keypad *
''      EF      Keypad /
''
''      F0      Left Shift
''      F1      Right Shift
''      F2      Left Ctrl
''      F3      Right Ctrl
''      F4      Left Alt
''      F5      Right Alt
''      F6      Left Win
''      F7      Right Win
''
''      FD      Scroll Lock State
''      FE      Caps Lock State
''      FF      Num Lock State
''
''      +100    if Shift
''      +200    if Ctrl
''      +400    if Alt
''      +800    if Win
''
''      eg. Ctrl-Alt-Delete = $6C9
''
''
'' Note: Driver will buffer up to 15 keystrokes, then ignore overflow.
#endif

CON

' *****************************************************************************
' *
' *     Terms of Use : MIT License
' *
' *****************************************************************************
'
' Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
' associated documentation files (the "Software"), to deal in the Software without restriction, including
' without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
' copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to
' the following conditions:
'
' The above copyright notice and this permission notice shall be included in all copies or substantial
' portions of the Software.
'
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
' LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
' NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
' WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

